'use strict';

/* For ESLint: List any global identifiers used in this file below */
/* global browser, ElemHideEmulation, splitSelector, logging, onReady,
runBandaids, parseUri, ElementTypes: true, FilterOptions: true */

let elemhide;

const typeMap = new Map([
  ['img', 'IMAGE'],
  ['input', 'IMAGE'],
  ['picture', 'IMAGE'],
  ['audio', 'MEDIA'],
  ['video', 'MEDIA'],
  ['frame', 'SUBDOCUMENT'],
  ['iframe', 'SUBDOCUMENT'],
  ['object', 'OBJECT'],
  ['embed', 'OBJECT'],
]);


// below are the entries from filtering/filteroptions.js
// they should be kept in sync.
//
ElementTypes = {
  NONE: 0,
  script: 1,
  image: 2,
  background: 4,
  stylesheet: 8,
  object: 16,
  subdocument: 32,
  object_subrequest: 64,
  media: 128,
  other: 256,
  xmlhttprequest: 512,
  document: 1024,
  elemhide: 2048,
  popup: 4096,

  // If you add something here, update .DEFAULTTYPES
};

// The types that are implied by a filter that doesn't explicitly specify types
ElementTypes.DEFAULTTYPES = 1023;

ElementTypes.FireFoxElementTypes = {
  TYPE_OTHER: 1,
  TYPE_SCRIPT: 2,
  TYPE_IMAGE: 3,
  TYPE_CSSIMAGE: 31, // Custom type
  TYPE_FAVICON: 32, // Custom type
  TYPE_STYLESHEET: 4,
  TYPE_OBJECT: 5,
  TYPE_DOCUMENT: 6,
  TYPE_SUBDOCUMENT: 7,
  TYPE_REDIRECT: 71, // Custom type
  TYPE_REFRESH: 8, // Unused
  TYPE_XBL: 9, // Unused
  TYPE_PING: 10, // Unused
  TYPE_XMLHTTPREQUEST: 11,
  TYPE_OBJECT_SUBREQUEST: 12,
};

ElementTypes.convertFireFoxContentType = function convertContentType(contentType) {
  switch (contentType) {
    case ElementTypes.FireFoxElementTypes.TYPE_OTHER:
      return ElementTypes.other;
    case ElementTypes.FireFoxElementTypes.TYPE_SCRIPT:
      return ElementTypes.script;
    case ElementTypes.FireFoxElementTypes.TYPE_IMAGE:
      return ElementTypes.image;
    case ElementTypes.FireFoxElementTypes.TYPE_CSSIMAGE:
      return ElementTypes.image;
    case ElementTypes.FireFoxElementTypes.TYPE_FAVICON:
      return ElementTypes.image;
    case ElementTypes.FireFoxElementTypes.TYPE_STYLESHEET:
      return ElementTypes.stylesheet;
    case ElementTypes.FireFoxElementTypes.TYPE_OBJECT:
      return ElementTypes.object;
    case ElementTypes.FireFoxElementTypes.TYPE_DOCUMENT:
      return ElementTypes.document;
    case ElementTypes.FireFoxElementTypes.TYPE_SUBDOCUMENT:
      return ElementTypes.subdocument;
    case ElementTypes.FireFoxElementTypes.TYPE_REDIRECT:
      return ElementTypes.NONE;
    case ElementTypes.FireFoxElementTypes.TYPE_REFRESH:
      return ElementTypes.NONE;
    case ElementTypes.FireFoxElementTypes.TYPE_XBL:
      return ElementTypes.NONE;
    case ElementTypes.FireFoxElementTypes.TYPE_PING:
      return ElementTypes.NONE;
    case ElementTypes.FireFoxElementTypes.TYPE_XMLHTTPREQUEST:
      return ElementTypes.xmlhttprequest;
    case ElementTypes.FireFoxElementTypes.TYPE_OBJECT_SUBREQUEST:
      return ElementTypes.object_subrequest;
    default:
      return ElementTypes.NONE;
  }
};

FilterOptions = {
  NONE: 0,
  THIRDPARTY: 1,
  MATCHCASE: 2,
  FIRSTPARTY: 4,
};

function getURLsFromObjectElement(element) {
  const url = element.getAttribute('data');
  if (url) {
    return [url];
  }

  for (const child of element.children) {
    if (child.localName === 'param') {
      const name = child.getAttribute('name');
      if (name === 'movie' // Adobe Flash
          || name === 'source' // Silverlight
          || name === 'src' // Real Media + Quicktime
          || name === 'FileName') { // Windows Media
        const value = child.getAttribute('value');
        if (value) {
          return [value];
        }
      }
    }
  }

  return [];
}

function getURLsFromAttributes(element) {
  const urls = [];

  if (element.src) {
    urls.push(element.src);
  }

  if (element.srcset) {
    for (const candidate of element.srcset.split(',')) {
      const url = candidate.trim().replace(/\s+\S+$/, '');
      if (url) {
        urls.push(url);
      }
    }
  }

  return urls;
}

function getURLsFromMediaElement(element) {
  const urls = getURLsFromAttributes(element);

  for (const child of element.children) {
    if (child.localName === 'source' || child.localName === 'track') {
      urls.push(...getURLsFromAttributes(child));
    }
  }

  if (element.poster) {
    urls.push(element.poster);
  }

  return urls;
}

function getURLsFromElement(element) {
  let urls;
  switch (element.localName) {
    case 'object':
      urls = getURLsFromObjectElement(element);
      break;

    case 'video':
    case 'audio':
    case 'picture':
      urls = getURLsFromMediaElement(element);
      break;

    default:
      urls = getURLsFromAttributes(element);
      break;
  }

  for (let i = 0; i < urls.length; i += 1) {
    if (/^(?!https?:)[\w-]+:/i.test(urls[i])) {
      urls.splice(i -= 1, 1);
    }
  }

  return urls;
}

function hideElement(element) {
  function doHide() {
    let propertyName = 'display';
    let propertyValue = 'none';
    if (element.localName === 'frame') {
      propertyName = 'visibility';
      propertyValue = 'hidden';
    }

    if (element.style.getPropertyValue(propertyName) !== propertyValue
        || element.style.getPropertyPriority(propertyName) !== 'important') {
      element.style.setProperty(propertyName, propertyValue, 'important');
    }
    if (element) {
      browser.runtime.sendMessage({ message: 'recordOneAdBlocked' });
    }
  }

  doHide();

  new MutationObserver(doHide).observe(element, {
    attributes: true,
    attributeFilter: ['style'],
  });
}

/* eslint-disable */
function checkCollapse(element) {
  const mediatype = typeMap.get(element.localName);
  if (!mediatype) {
    return;
  }

  const urls = getURLsFromElement(element);
  if (urls.length === 0) {}
//
// TODO
//  ext.backgroundPage.sendMessage(
//    {
//      type: "filters.collapse",
//      urls,
//      mediatype,
//      baseURL: document.location.href
//    },
//
//    collapse =>
//    {
//      if (collapse)
//      {
//        hideElement(element);
//      }
//    }
//  );
}
/* eslint-enable */

function ElemHide(data) {
  this.shadow = this.createShadowTree();
  this.styles = new Map();
  this.inline = true;
  this.localData = data;

  this.elemHideEmulation = new ElemHideEmulation(
    this.addSelectors.bind(this),
    this.hideElements.bind(this),
  );
}
ElemHide.prototype = {
  selectorGroupSize: 200,

  createShadowTree() {
    // Use Shadow DOM if available as to not mess with with web pages that
    // rely on the order of their own <style> tags (#309). However, creating
    // a shadow root breaks running CSS transitions. So we have to create
    // the shadow root before transistions might start (#452).
    if (!('createShadowRoot' in document.documentElement)) {
      return null;
    }

    // Both Firefox and Chrome 66+ support user style sheets, so we can avoid
    // creating an unnecessary shadow root on these platforms.
    const match = /\bChrome\/(\d+)/.exec(navigator.userAgent);
    if (!match || match[1] >= 66) {
      return null;
    }

    // Using shadow DOM causes issues on some Google websites,
    // including Google Docs, Gmail and Blogger (#1770, #2602, #2687).
    if (/\.(?:google|blogger)\.com$/.test(document.domain)) {
      return null;
    }

    // Finally since some users have both AdBlock and Adblock Plus installed we
    // have to consider how the two extensions interact. For example we want to
    // avoid creating the shadowRoot twice.
    const shadow = document.documentElement.shadowRoot
                || document.documentElement.createShadowRoot();
    shadow.appendChild(document.createElement('content'));

    return shadow;
  },

  addSelectorsInline(selectors, groupName, appendOnly = false) {
    let style = this.styles.get(groupName);

    if (style && !appendOnly) {
      while (style.sheet.cssRules.length > 0) {
        style.sheet.deleteRule(0);
      }
    }

    if (selectors.length === 0) {
      return;
    }

    if (!style) {
      // Create <style> element lazily, only if we add styles. Add it to
      // the shadow DOM if possible. Otherwise fallback to the <head> or
      // <html> element. If we have injected a style element before that
      // has been removed (the sheet property is null), create a new one.
      style = document.createElement('style');
      (this.shadow || document.head
                   || document.documentElement).appendChild(style);

      // It can happen that the frame already navigated to a different
      // document while we were waiting for the background page to respond.
      // In that case the sheet property will stay null, after addind the
      // <style> element to the shadow DOM.
      if (!style.sheet) {
        return;
      }

      this.styles.set(groupName, style);
    }

    // If using shadow DOM, we have to add the ::content pseudo-element
    // before each selector, in order to match elements within the
    // insertion point.
    let preparedSelectors = [];
    if (this.shadow) {
      for (const selector of selectors) {
        const subSelectors = splitSelector(selector);
        for (const subSelector of subSelectors) {
          preparedSelectors.push(`::content ${subSelector}`);
        }
      }
    } else {
      preparedSelectors = selectors;
    }

    // Chromium's Blink engine supports only up to 8,192 simple selectors, and
    // even fewer compound selectors, in a rule. The exact number of selectors
    // that would work depends on their sizes (e.g. "#foo .bar" has a
    // size of 2). Since we don't know the sizes of the selectors here, we
    // simply split them into groups of 1,024, based on the reasonable
    // assumption that the average selector won't have a size greater than 8.
    // The alternative would be to calculate the sizes of the selectors and
    // divide them up accordingly, but this approach is more efficient and has
    // worked well in practice. In theory this could still lead to some
    // selectors not working on Chromium, but it is highly unlikely.
    // See issue #6298 and https://crbug.com/804179
    for (let i = 0; i < preparedSelectors.length; i += this.selectorGroupSize) {
      const selector = preparedSelectors.slice(i, i + this.selectorGroupSize).join(', ');
      style.sheet.insertRule(
        `${selector}{display: none !important;}`,
        style.sheet.cssRules.length,
      );
    }
  },

  addSelectors(selectors, filters, groupName = 'emulated', appendOnly = false) {
    if (this.inline || this.inlineEmulated) {
      // Insert the style rules inline if we have been instructed by the
      // background page to do so. This is usually the case, except on platforms
      // that do support user stylesheets via the browser.tabs.insertCSS API
      // (Firefox 53 onwards for now and possibly Chrome in the near future).
      // Once all supported platforms have implemented this API, we can remove
      // the code below. See issue #5090.
      // Related Chrome and Firefox issues:
      // https://bugs.chromium.org/p/chromium/issues/detail?id=632009
      // https://bugzilla.mozilla.org/show_bug.cgi?id=1310026
      this.addSelectorsInline(selectors, groupName, appendOnly);
    }
  },

  hideElements(elements) {
    elements.forEach((element) => {
      hideElement(element);
    });
  },

  apply() {
    if (this.localData && this.localData.inline) {
      this.inline = this.localData.inline;
    }

    if (this.inline && this.localData && this.localData.selectors) {
      this.addSelectorsInline(this.localData.selectors, 'standard');
    }
    if (this.localData && this.localData.advanceSelectors) {
      this.elemHideEmulation.apply(this.localData.advanceSelectors);
    }
  },
};

function runInPageContext(fn, arg) {
  const script = document.createElement('script');
  script.type = 'application/javascript';
  script.async = false;
  script.textContent = `(${fn})(${JSON.stringify(arg)});`;
  document.documentElement.appendChild(script);
  document.documentElement.removeChild(script);
}

// cache a reference to window.confirm
// so that web sites can not clobber the default implementation
const abConfirm = window.confirm;

// Return the ElementType element type of the given element.
function typeForElement(el) {
  // TODO: handle background images that aren't just the BODY.
  switch (el.nodeName.toUpperCase()) {
    case 'INPUT':
    case 'IMG':
      return ElementTypes.image;
    case 'SCRIPT':
      return ElementTypes.script;
    case 'OBJECT':
    case 'EMBED':
      return ElementTypes.object;
    case 'VIDEO':
    case 'AUDIO':
    case 'SOURCE':
      return ElementTypes.media;
    case 'FRAME':
    case 'IFRAME':
      return ElementTypes.subdocument;
    case 'LINK':

      // favicons are reported as 'other' by onBeforeRequest.
      // if this is changed, we should update this too.
      if (/(^|\s)icon($|\s)/i.test(el.rel)) {
        return ElementTypes.other;
      }
      return ElementTypes.stylesheet;
    case 'BODY':
      return ElementTypes.background;
    default:
      return ElementTypes.NONE;
  }
}

// If url is relative, convert to absolute.
function relativeToAbsoluteUrl(url) {
  // Author: Tom Joseph of AdThwart

  if (!url) {
    return url;
  }

  // If URL is already absolute, don't mess with it
  if (/^[a-zA-Z-]+:/.test(url)) {
    return url;
  }

  if (url[0] === '/') {
    // Leading // means only the protocol is missing
    if (url[1] && url[1] === '/') {
      return document.location.protocol + url;
    }

    // Leading / means absolute path
    return `${document.location.protocol}//${document.location.host}${url}`;
  }

  // Remove filename and add relative URL to it
  const base = document.baseURI.match(/.+\//);
  if (!base) {
    return `${document.baseURI}/${url}`;
  }
  return base[0] + url;
}

// Do not make the frame display a white area
// Not calling .remove(); as this causes some sites to reload continuesly
function removeFrame(element) {
  let el = element;
  const parentEl = el.parentNode;
  const cols = ((parentEl.getAttribute('cols') || '').indexOf(',') > 0);
  if (!cols && (parentEl.getAttribute('rows') || '').indexOf(',') <= 0) {
    return;
  }

  // Figure out which column or row to hide
  let index = 0;
  while (el.previousElementSibling) {
    index += 1;
    el = el.previousElementSibling;
  }

  // Convert e.g. '40,20,10,10,10,10' into '40,20,10,0,10,10'
  const attr = (cols ? 'cols' : 'rows');
  const sizes = parentEl.getAttribute(attr).split(',');
  sizes[index] = '0';
  parentEl.setAttribute(attr, sizes.join(','));
}

// Remove an element from the page.
function destroyElement(el, elType) {
  if (el.nodeName === 'FRAME') {
    removeFrame(el);
  } else if (elType !== ElementTypes.script) {
    // There probably won't be many sites that modify all of these.
    // However, if we get issues, we might have to set the location and size
    // via the css properties position, left, top, width and height
    el.style.setProperty('display', 'none', 'important');
    el.style.setProperty('visibility', 'hidden', 'important');
    el.style.setProperty('opacity', '0', 'important');
    const w = (el.width === undefined ? -1 : el.width);
    const h = (el.height === undefined ? -1 : el.height);
    el.style.setProperty('background-position', `${w}px ${h}px`);
    el.setAttribute('width', 0);
    el.setAttribute('height', 0);
  }
}

function debugPrintSelectorMatches(selectors) {
  selectors
    .filter(selector => document.querySelector(selector))
    .forEach((selector) => {
      let matches = '';
      const elems = document.querySelectorAll(selector);
      for (let i = 0; i < elems.length; i += 1) {
        const el = elems[i];
        matches += `        ${el.nodeName}#${el.id}.${el.className}\n`;
      }

      const opts = { selector: `##${selector}`, matches };
      browser.runtime.sendMessage({ message: 'debug_report_elemhide', opts });
    });
}

function handleABPLinkClicks() {
  // Subscribe to the list when you click an abp: link
  const elems = document.querySelectorAll('[href^="abp:"], [href^="ABP:"]');
  const abplinkhandler = function (event) {
    if (event.isTrusted === false) {
      return;
    }

    event.preventDefault();
    const searchquery = this.href.replace(/^.+?\?/, '?');
    if (searchquery) {
      const queryparts = parseUri.parseSearch(searchquery);
      const url = queryparts.location;
      const { title } = queryparts;
      browser.runtime.sendMessage({
        message: 'open.options.page.subscriptions.add',
        title,
        url,
      });
    }
  };

  for (let i = 0; i < elems.length; i += 1) {
    elems[i].addEventListener('click', abplinkhandler, false);
  }
}

// Called at document load.
// inputs:
//   startPurger: function to start watching for elements to remove.
//   stopPurger: function to stop watch for elemenst to remove, called in case
//               AdBlock should not be running.
//   success?: function called at the end if AdBlock should run on the page.
function adblockBegin(inputs) {
  if (document.location.href === 'about:blank') { // Firefox does this
    return;
  }
  if (!(document.documentElement instanceof HTMLElement)) {
    return;
  } // Only run on HTML pages

  inputs.startPurger();

  const opts = { domain: document.location.hostname };

  browser.runtime.sendMessage({ message: 'get_content_script_data', opts }, function dataCallback(data) {
    if (data && data.settings && data.settings.debug_logging) {
      logging(this, true);
    }

    elemhide = new ElemHide(data);
    elemhide.apply();

    document.addEventListener('error', (event) => {
      checkCollapse(event.target);
    }, true);

    document.addEventListener('load', (event) => {
      const element = event.target;
      if (/^i?frame$/.test(element.localName)) {
        checkCollapse(element);
      }
    }, true);

    if (data && !data.running) {
      inputs.stopPurger();
      return;
    }

    onReady(() => {
      // TODO: ResourceList could pull html.innerText from page instead:
      // we could axe this
      if (data && data.settings && data.settings.debug_logging) {
        debugPrintSelectorMatches(data.selectors || []);
      }

      if (typeof runBandaids === 'function') {
        runBandaids('new');
      }

      handleABPLinkClicks();
    });
  });
}
